#pragma once
#include <vector>
#include <string>
#include "yubikey.h"
#include "VectorString.h"
#include "SplitString.h"
#include <algorithm>


template <typename S>
void ToLower(S& s)
{
	std::transform(s.begin(), s.end(), s.begin(), ::tolower);
}

template<typename NARROWSTRING, typename VECTOR>
VECTOR ModhexDecodeN(const NARROWSTRING& otp_entry_c)
{
	NARROWSTRING otp_entry(otp_entry_c);
	if (otp_entry.length() == 0)
	{
		SUPERLOG(_T("ModHex input is empty"));
		throw (DWORD)ERROR_INVALID_PARAMETER;
	}

	// even number of characters
	if (otp_entry.length() % 2 != 0)
	{
		SUPERLOG(_T("ModHex input has an odd numer of characters (%d)"), otp_entry.length());
		throw (DWORD)ERROR_INVALID_PARAMETER;
	}
	ToLower(otp_entry);

	VECTOR decoded(otp_entry.length() / 2);
	// take the first possible layout (orders by preference of qwerty, dvorak, colemak)
	ModHexClass layout = yubikey_modhex_serialize(yubikey_modhex_classify((const unsigned char*)(otp_entry.c_str()), otp_entry.length())).front();
	if (layout != ModHexClass::NotModHex)
	{
		SUPERLOG(_T("It looks like modhex layout(%d)"), layout);
		yubikey_modhex_decode(&decoded[0], (const unsigned char*)(otp_entry.c_str()), decoded.size(), layout);
	}
	else
	{
		SUPERLOG(_T("Has non modhex chars"));
		throw (DWORD)ERROR_INVALID_PARAMETER;
	}
	return decoded;

}

template<typename S, typename C>
std::pair<S, S> SplitFieldOnFirstDash(const S& input)
{
	std::pair<S, S> res;
	std::list<S> split = SplitString<S,C>(input, _T('-'));
	if (split.size() < 2)
	{
		throw (DWORD)ERROR_INVALID_PARAMETER;
	}
	res.first = *(split.begin());
	// take everything right of the first dash, which means everything except the first split's length, and one (for the dash)
	res.second = S(input.begin() + res.first.length() + 1, input.end());
	return res;
}
template<typename S, typename C>
std::pair<S, S> SplitFieldOnLastDash(const S& input)
{
	std::pair<S, S> res;
	std::list<S> split = SplitString<S,C>(input, _T('-'));
	if (split.size() < 2)
	{
		throw (DWORD)ERROR_INVALID_PARAMETER;
	}
	res.second = *(split.rbegin());
	// take everything left of the last dash, which means everything except the last split's length, and one (for the dash)
	res.first = S(input.begin(), input.begin() + input.length() - res.second.length() - 1);
	return res;
}
template<typename S, typename C>
std::pair<S, S> SplitFieldOnSingleAt(const S& input)
{
	std::list<S> res = SplitString<S, C>(input, _T('@'));
	if (res.size() != 2)
	{
		throw (DWORD)ERROR_INVALID_PARAMETER;
	}
	return std::pair<S, S>(res.front(), res.back());
}
template<typename S>
bool ContainsOnlyDigits(const S& input)
{
	return S::npos == input.find_first_not_of(_T("0123456789"));
}

template<typename S, typename C>
void TrimRight(S& v, C c)
{
	if (v.length() > 0 && v.back() == c)
		v.pop_back();
}
template<typename S, typename C>
void TrimLeft(S& v, C c)
{
	if (v.length() > 0 && v.front() == c)
		v.erase(v.begin());
}

enum OTPType
{
	None,
	Yubikey,
	Oath
};

template <typename STRING, typename SECSTRING, typename SECVECTOR>
struct ParsedOTPDataT
{
	OTPType otp_type;
	STRING otp_string;
	STRING otp_string_short;
	STRING username;
	SECSTRING password;
	SECVECTOR otp_bytes;
};

template <typename STRING, typename CHAR>
STRING BuildOTPString(STRING username, STRING otp)
{
	STRING upnSuffix;
	try
	{
		std::pair<STRING, STRING> split = SplitFieldOnSingleAt<STRING, CHAR>(username);
		upnSuffix = split.second;
		username = split.first;
	}
	catch (DWORD e)
	{
		DEBUGLOG(_T("Username is not a UPN (%d)"), e);
	}
	STRING otp_string;
	otp_string = username;
	otp_string += _T("-");
	otp_string += otp;
	if (upnSuffix.size() > 0)
	{
		otp_string += _T("@");
		otp_string += upnSuffix;
	}
	return otp_string;
}
template <typename STRING, typename CHAR, typename NARROWSTRING, typename SECSTRING, typename VECTOR, typename SECVECTOR>bool ParseOTP(STRING username, const SECSTRING& password, ParsedOTPDataT<STRING, SECSTRING, SECVECTOR>& res, unsigned short expectedYubikeyOtpChars, unsigned short expectedOathOtpChars)
{
	SUPERLOG(_T("ParseOTP %s %s %d %d"), username.c_str(), password.c_str(), expectedYubikeyOtpChars, expectedOathOtpChars);
	res.otp_type = OTPType::None;
	res.username = username;
	res.password = password;

	if (username.length() > 0)
	{
		do // allow multiple breaks to end of block
		{
			STRING upnSuffix;
			try
			{
				// if this works, we're guaranteed that there's no other @ in the string
				std::pair<STRING, STRING> split = SplitFieldOnSingleAt<STRING,CHAR>(username);
				upnSuffix = split.second;
				username = split.first;
				// ok but what if they said pre@suffix-otp, I'd like to make that work too
				ParsedOTPDataT<STRING, SECSTRING, SECVECTOR> upnExtra;
				if (ParseOTP<STRING, CHAR, NARROWSTRING, SECSTRING, VECTOR, SECVECTOR>(upnSuffix, STRING(), upnExtra, expectedYubikeyOtpChars, expectedOathOtpChars))
				{
					STRING newu = split.first;
					newu += _T("-");
					newu += upnExtra.otp_string_short;
					DEBUGLOG(_T("Username (%s@%s) seems to be a UPN with OTP in the suffix section (%s) (%s), rewriting to %s-%s@%s"),
						username.c_str(), upnSuffix.c_str(), upnExtra.username.c_str(), upnExtra.otp_string_short.c_str(),
						newu.c_str(), upnExtra.otp_string_short.c_str(), upnExtra.username.c_str());
					username = newu;
					upnSuffix = upnExtra.username;
				}
			}
			catch (DWORD e)
			{
				DEBUGLOG(_T("Username is not a UPN (%d)"), e);
			}
			// note this abandons the ancient "just see if the entire string is exactly modhex chars" logic
			// and *demands* we have an expected yubikey length.  I think this will be fine because everything else
			// expects the length to be set everywhere else, and the default is correct for old installs.
			if (expectedYubikeyOtpChars != 0) try
			{
				if (username.length() < expectedYubikeyOtpChars) throw (DWORD)ERROR_INVALID_PARAMETER;

				try
				{
					TrimRight(username, _T('\t'));
					res.otp_bytes = ModhexDecode(SECSTRING(username.end() - expectedYubikeyOtpChars, username.end()));
					res.otp_type = OTPType::Yubikey;
					res.otp_string_short = STRING(username.end() - expectedYubikeyOtpChars, username.end());
					res.otp_string = res.otp_string_short;
					username = STRING(username.begin(), username.begin() + username.length() - expectedYubikeyOtpChars);
					//  ^ possibly a real username and possibly a dash
					if (username.length() > 0 && username.back() == _T('-'))
					{
						SUPERLOG(_T("Stripping a dash between username and otp"));
						TrimRight(username, _T('-'));
					}
					res.username = username; // may be ""
					if (username.size() > 0) // anything left?
					{
						// recall otp_string should be everything we need to do the lookup					
						res.otp_string = username;
						res.otp_string += _T("-");
						res.otp_string += res.otp_string_short;
						if (upnSuffix.size() > 0)
						{
							res.username += _T("@");
							res.username += upnSuffix;
							res.otp_string += _T("@");
							res.otp_string += upnSuffix;
						}
					}

					DEBUGLOG(_T("It looks like a modhex (username-?)OTP in the username field (%s)(%s)(%s)"), res.username.c_str(), res.otp_string_short.c_str(), res.otp_string.c_str());
					break; /* to password block */
				}
				catch (DWORD e)
				{
					res.otp_bytes = ModhexDecode(SECSTRING(username.begin(), username.begin() + expectedYubikeyOtpChars));
					res.otp_type = OTPType::Yubikey;
					res.otp_string_short = STRING(username.begin(), username.begin() + expectedYubikeyOtpChars);
					res.otp_string = res.otp_string_short;
					username = STRING(username.begin() + expectedYubikeyOtpChars, username.end());
					// ^ possibly a \t followed by possibly a dash followed by possibly a real username
					TrimLeft(username, _T('\t'));
					if (username.length() > 0 && username.front() == _T('-'))
					{
						SUPERLOG(_T("Stripping a dash between otp and username"));
						TrimLeft(username, _T('-'));
					}
					res.username = username; // may be ""
					if (username.size() > 0) // anything left?
					{
						// recall otp_string should be everything we need to do the lookup					
						res.otp_string = username;
						res.otp_string += _T("-");
						res.otp_string += res.otp_string_short;
						if (upnSuffix.size() > 0)
						{
							res.username += _T("@");
							res.username += upnSuffix;
							res.otp_string += _T("@");
							res.otp_string += upnSuffix;
						}
					}
					
					DEBUGLOG(_T("It looks like a modhex OTP(-username?) in the username field (%s)(%s)(%s)"), res.username.c_str(), res.otp_string_short.c_str(), res.otp_string.c_str());
					break; /* to password block */
				}
			}
			catch (DWORD e)
			{
				DEBUGLOG(_T("Username does not contain a modhex otp (%d)"), e);
			}
			// we need to have some expectation, not just "try whatever and see if it works"
			// not because we can't, but because chars-## is a plausible username format
			// we could write this to expect a bracketed range, if that becomes helpful
			if (expectedOathOtpChars > 0)
			{
				// special case rejection
				// https://tix.authlite.com/tickets/5KG6D14YMNSJ
				if (username.length() == 55 && std::wstring::npos == username.find_first_not_of(_T("0123456789-")))
				{
					DEBUGLOG(_T("Username could be a bitlocker recovery key and is super not likely to be a username-otp, so skipping"));
					break; // to password block
				}
				// https://tix.authlite.com/tickets/5YGSBKCM98HJ
				// https://stackoverflow.com/questions/1878001/how-do-i-check-if-a-c-stdstring-starts-with-a-certain-string-and-convert-a
				if (0 == username.rfind(_T("S-1-"), 0) && std::wstring::npos == username.find_first_not_of(_T("S0123456789-")))
				{
					DEBUGLOG(_T("Username appears to be a SID and is super not likely to be a username-otp, so skipping"));
					break; // to password block
				}
				// https://tix.authlite.com/tickets/0J5WCBRYG4F8
				if ((0 == username.rfind(_T("SMTP:"), 0)) ||
					(0 == username.rfind(_T("smtp:"), 0)))
				{
					DEBUGLOG(_T("Username starts with 'SMTP: or smtp:', so skipping")); // l-case 1G1EMN5QTY4R
					break; // to password block
				}
				try
				{
					std::pair<STRING, STRING> split = SplitFieldOnLastDash<STRING, CHAR>(username);
					if (split.second.length() != expectedOathOtpChars)
						throw (DWORD)ERROR_INVALID_PARAMETER;
					if (!ContainsOnlyDigits<STRING>(split.second))
						throw (DWORD)ERROR_INVALID_PARAMETER;
					DEBUGLOG(_T("It looks like username- OATH OTP"));
					res.otp_type = OTPType::Oath;
					res.otp_string_short = split.second;
					res.username = split.first;
					res.otp_string = username; // which has the -otp on it
					if (upnSuffix.size() > 0)
					{
						res.username += _T("@");
						res.username += upnSuffix;
						res.otp_string += _T("@");
						res.otp_string += upnSuffix;
					}
					res.otp_bytes = TxVector<VECTOR, SECVECTOR>(VectorFromString<STRING, VECTOR, CHAR>(username));
					res.password = password;
					break; /* to password block */
				}
				catch (DWORD e)
				{
					/*turning off "OATH-username" as dangerous and never needed, some customers use 6 digit usernames
					try
					{
						std::pair<STRING, STRING> split = SplitFieldOnFirstDash<STRING, CHAR>(username);
						if (split.first.length() != expectedOathOtpChars)
							throw (DWORD)ERROR_INVALID_PARAMETER;
						if (!ContainsOnlyDigits<STRING>(split.first))
							throw (DWORD)ERROR_INVALID_PARAMETER;
						DEBUGLOG(_T("It looks like OATH OTP- username"));
						res.otp_type = OTPType::Oath;
						res.otp_string_short = split.first;
						res.username = split.second;
						//normalize order res.otp_string = username; // which has otp- on it
						res.otp_string = res.username;
						res.otp_string += _T("-");
						res.otp_string += res.otp_string_short;
						if (upnSuffix.size() > 0)
						{
							res.username += _T("@");
							res.username += upnSuffix;
							res.otp_string += _T("@");
							res.otp_string += upnSuffix;
						}
						res.otp_bytes = TxVector<VECTOR, SECVECTOR>(VectorFromString<STRING, VECTOR, CHAR>(username));
						res.password = password;
						break; /* to password block * /
					}
					catch (DWORD e)
					{*/
						DEBUGLOG(_T("Username is not in OATH format (%d)"), e);
					//}
				}
			}
		} while (false); //never loop
	} // if username

	// res.username may have been set above. check for *another* otp in password field, in case the caller remembers old read-only usernames
	if (password.length() > 0) // don't bother unless there's something to look at
	{
		// tho leaving username alone for the most part, I actually still
		// need to do this, because I have to assemble the otp_string
		// use the pre-chewed value from above (in case there was an OTP in there)
		username = res.username;
		STRING upnSuffix;
		try
		{
			std::pair<STRING, STRING> split = SplitFieldOnSingleAt<STRING, CHAR>(username);
			upnSuffix = split.second;
			username = split.first;
		}
		catch (DWORD e)
		{
			DEBUGLOG(_T("Username is not a UPN (%d)"), e);
		}

		// can only do password modhex detection if we know the exact otp string size because there's no separator char to look for
		if (expectedYubikeyOtpChars != 0) try
		{
			if (password.length() < expectedYubikeyOtpChars) throw (DWORD)ERROR_INVALID_PARAMETER;

			SECSTRING passtemp = password;
			try
			{
				TrimRight(passtemp, _T('\t'));
				res.otp_bytes = ModhexDecode(SECSTRING(passtemp.end() - expectedYubikeyOtpChars, passtemp.end()));
				res.otp_type = OTPType::Yubikey;
				res.otp_string_short = STRING(passtemp.end() - expectedYubikeyOtpChars, passtemp.end());
				res.otp_string = res.otp_string_short;
				if (!username.empty())
				{
					res.username = username;
					res.otp_string = username;
					res.otp_string += _T("-");
					res.otp_string += res.otp_string_short;
					if (upnSuffix.size() > 0)
					{
						res.username += _T("@");
						res.username += upnSuffix;
						res.otp_string += _T("@");
						res.otp_string += upnSuffix;
					}
				}
				passtemp = SECSTRING(passtemp.begin(), passtemp.begin() + passtemp.length() - expectedYubikeyOtpChars);
				if (passtemp.length() > 0 && passtemp.back() == _T('-'))
				{
					SUPERLOG(_T("Stripping a dash between password and otp"));
					TrimRight(passtemp, _T('-'));
				}
				res.password = passtemp;
				DEBUGLOG(_T("It looks like a modhex pass+otp in the password field"));
				return true;
			}
			catch (DWORD e)
			{
				res.otp_bytes = ModhexDecode(SECSTRING(passtemp.begin(), passtemp.begin() + expectedYubikeyOtpChars));
				res.otp_type = OTPType::Yubikey;
				res.otp_string_short = STRING(passtemp.begin(), passtemp.begin() + expectedYubikeyOtpChars);
				res.otp_string = res.otp_string_short;
				if (!username.empty())
				{
					res.username = username;
					res.otp_string = username;
					res.otp_string += _T("-");
					res.otp_string += res.otp_string_short;
					if (upnSuffix.size() > 0)
					{
						res.username += _T("@");
						res.username += upnSuffix;
						res.otp_string += _T("@");
						res.otp_string += upnSuffix;
					}
				}
				passtemp = SECSTRING(passtemp.begin() + expectedYubikeyOtpChars, passtemp.end());
				TrimLeft(passtemp, _T('\t'));
				if (passtemp.length() > 0 && passtemp.front() == _T('-'))
				{
					SUPERLOG(_T("Stripping a dash between otp and password"));
					TrimLeft(passtemp, _T('-'));
				}
				res.password = passtemp;
				DEBUGLOG(_T("It looks like a modhex OTP+pass in the password field"));
				return true;
			}
		}
		catch (DWORD e)
		{
			DEBUGLOG(_T("password does not have a modhex otp on either side(%d)"), e);
		}

		if (expectedOathOtpChars > 0) try
		{
			std::pair<SECSTRING, SECSTRING> split = SplitFieldOnLastDash<SECSTRING, CHAR>(password);
			if (split.second.length() != expectedOathOtpChars)
				throw (DWORD)ERROR_INVALID_PARAMETER;
			if (!ContainsOnlyDigits<SECSTRING>(split.second))
				throw (DWORD)ERROR_INVALID_PARAMETER;
			DEBUGLOG(_T("It looks like password- OATH OTP"));
			res.otp_type = OTPType::Oath;
			res.otp_string_short = TxVector<SECSTRING, STRING>(split.second);
			res.otp_string = res.otp_string_short;
			if (!username.empty())
			{
				res.username = username;
				res.otp_string = username;
				res.otp_string += _T("-");
				res.otp_string += res.otp_string_short;
				if (upnSuffix.size() > 0)
				{
					res.username += _T("@");
					res.username += upnSuffix;
					res.otp_string += _T("@");
					res.otp_string += upnSuffix;
				}
			}
			res.otp_bytes = TxVector<VECTOR, SECVECTOR>(VectorFromString<STRING, VECTOR, CHAR>(res.otp_string));
			res.password = split.first;
			return true;
		}
		catch (DWORD e)
		{
			try
			{
				std::pair<SECSTRING, SECSTRING> split = SplitFieldOnFirstDash<SECSTRING, CHAR>(password);
				if (split.first.length() != expectedOathOtpChars)
					throw (DWORD)ERROR_INVALID_PARAMETER;
				if (!ContainsOnlyDigits<SECSTRING>(split.first))
					throw (DWORD)ERROR_INVALID_PARAMETER;
				DEBUGLOG(_T("It looks like OATH OTP- password"));
				res.otp_type = OTPType::Oath;
				res.otp_string_short = TxVector<SECSTRING, STRING>(split.first);
				res.otp_string = res.otp_string_short;
				if (!username.empty())
				{
					res.username = username;
					res.otp_string = username;
					res.otp_string += _T("-");
					res.otp_string += res.otp_string_short;
					if (upnSuffix.size() > 0)
					{
						res.username += _T("@");
						res.username += upnSuffix;
						res.otp_string += _T("@");
						res.otp_string += upnSuffix;
					}
				}
				res.otp_bytes = TxVector<VECTOR, SECVECTOR>(VectorFromString<STRING, VECTOR, CHAR>(res.otp_string));
				res.password = split.second;
				return true;
			}
			catch (DWORD e)
			{
				DEBUGLOG(_T("password does not have a dashed OATH otp on either side (%d)"), e);
			}
		}
	} // if password > 0 len
	else
		DEBUGLOG(_T("Skipping password check (empty)"));

	return res.otp_type != OTPType::None;

}
template <typename STRING, typename CHAR>
bool IsOTPish(STRING s)
{
	// strip potential UPN suffix as irrelevant
	try
	{
		std::pair<STRING, STRING> split = SplitFieldOnSingleAt<STRING, CHAR>(s);
		s = split.first;
	}
	catch (DWORD e){}
	
	// needs to handle varieties 
	// something-(modhex)
	// something-(\d{6+})
	// {modhex}-something
	// {modhex}
	size_t minyk = (YUBIKEY_BLOCK_SIZE * sizeof(CHAR));
	try
	{
		std::pair<STRING, STRING> split = SplitFieldOnLastDash<STRING, CHAR>(s);//or throw
		if (ContainsOnlyDigits<STRING>(split.second) && split.second.length() >= 6)
		{
			return true;
		}
		if (split.second.length() >= minyk)
		{
			ModhexDecode(split.second);//or throw
			return true;
		}
	}
	catch(DWORD){}
	try
	{
		std::pair<STRING, STRING> split = SplitFieldOnFirstDash<STRING, CHAR>(s);//or throw
		if (split.first.length() >= minyk)
		{
			ModhexDecode(split.first);//or throw
			return true;
		}
	}
	catch(DWORD){}
	try
	{
		if (s.length() >= minyk)
		{
			ModhexDecode(s);//or throw
			return true;
		}
	}
	catch (DWORD){}
	return false;
}
